from pydantic import (
    AnyUrl,
)

from .base_model import BaseModel


class BootfileData(BaseModel):
    """
    Contains information on a bootfile (such as grub) as well as which
    PXE client architecture it applies to. For more information on PXE
    client architectures, see https://tools.ietf.org/html/rfc4578#section-2.1
    """
    client_arch: int
    image_source_url: AnyUrl
    image_sha256sum: str

    def get_path(self):
        """
        Returns the path to the file. This may be used by TFTP packaged
        deploy to determine where to save the file as well as by the
        DHCP packaged deploy to determine what value to provide in
        DHCP Option 67 (Bootfile-Name)
        """
        return self.image_source_url.path.lstrip('/')



from ipaddress import (
    AddressValueError,
    IPv4Interface,
)


# ==========
# EXCEPTIONS
# ==========

class InvalidIPAddressError(ValueError):

    def __init__(self, ip_address):
        msg = f"'{ip_address}' must use CIDR notation."
        super().__init__(msg)


class UnsupportedSchemaVersion(ValueError):

    def __init__(self, data_obj, supported_schema_versions):
        msg = f"{str(type(data_obj))} is not supported. Supported schemas are " \
              f"{supported_schema_versions}"
        super().__init__(msg)


# ==========
# VALIDATORS
# ==========

def validate_cidr_notation(ip_address):
    try:
        IPv4Interface(ip_address)
    except AddressValueError:
        raise InvalidIPAddressError(ip_address)


def validate_schema_version(data, supported_schema_versions):
    if type(data) not in supported_schema_versions:
        raise UnsupportedSchemaVersion(
            data,
            supported_schema_versions
        )



from typing import (
    List,
    Optional,
)

from .base_model import BaseModel
from .dhcp_data import DhcpData
from .dns_data import DnsData
from .tftp_data import TftpData


class DnsmasqData(BaseModel):
    """
    Contains information on which interfaces dnsmasq should listen to for client
    requests. Note that all fields are optional. If the interfaces field is not
    assigned, then it will be assumed that dnsmasq should listen on all interfaces.
    Likewise if dhcp is not assigned, then it is assumed that dnsmasq should not
    serve DHCP requests.
    """
    interfaces: List[str] = []
    dhcp: Optional[DhcpData]
    dns: Optional[DnsData]
    tftp: Optional[TftpData]



from pathlib import Path

from pyinfra.api import (
    deploy,
)
from pyinfra.operations import (
    files,
    server,
)

from aircraft.deploys.synology.models import v1beta2
from aircraft.validators import validate_schema_version

# This is bad. Each deploy dir should be independent. I'm just
# too lazy to copy the templates and files right now because it
# would mean modifying two copies at the same time. Some fix is
# needed down the road.
deploy_dir = Path(__file__).parent.parent / 'ubuntu'


@deploy('Configure the PXE server')
def configure(state=None, host=None):
    supported_schema_versions = [
        v1beta2.PxeData,
    ]

    validate_schema_version(host.data.pxe, supported_schema_versions)

    templates_base = deploy_dir / 'templates'
    # Make sure to strip any trailing / in os_image_source_url.path
    # otherwise the whole thing will resolve into an absolute path
    # starting only with os_image_source_url.path.
    downloaded_iso_ssh_path = \
        host.data.pxe.http.root_dir / host.data.pxe.os_image_source_url.path.lstrip('/')

    files.directory(
        name='Ensure OS image directory',
        path=str(host.data.pxe.http.root_dir),
        present=True,

        host=host, state=state,
    )

    os_image = files.download(
        name='Download OS image',
        src=str(host.data.pxe.os_image_source_url),
        dest=str(downloaded_iso_ssh_path),
        sha256sum=host.data.pxe.os_image_sha256sum,
        sudo=True,

        host=host, state=state,
    )

    kernel_path = str(host.data.pxe.tftp.root_dir / 'vmlinuz')
    initrd_path = str(host.data.pxe.tftp.root_dir / 'initrd')

    if host.fact.file(kernel_path) is None or \
       host.fact.file(initrd_path) is None or \
       os_image.changed:
        server.shell(
            name='Mount the ISO to /mnt',
            commands=[
                f'mount | grep "{downloaded_iso_ssh_path} on /mnt" || '
                f'mount {downloaded_iso_ssh_path} /mnt',
            ],
            sudo=True,

            host=host, state=state
        )

        server.shell(
            name="Extract kernel and ramdisk image from ISO",
            commands=[
                f'cp /mnt/casper/vmlinuz {kernel_path}',
                f'cp /mnt/casper/initrd {initrd_path}',
            ],
            sudo=True,

            host=host, state=state,
        )

        server.shell(
            name='Unmount the ISO',
            commands=[
                'umount /mnt',
            ],
            sudo=True,

            host=host, state=state
        )

    for bootfile in host.data.pxe.bootfiles:

        bootfile_ssh_dir = (host.data.pxe.tftp.root_dir / bootfile.get_path()).parent

        files.directory(
            name='Ensure bootfile directory',
            path=str(bootfile_ssh_dir),
            present=True,

            host=host, state=state,
        )

        files.download(
            name=f'Download bootfile {bootfile.image_source_url}',
            src=str(bootfile.image_source_url),
            dest=str(host.data.pxe.tftp.root_dir / bootfile.get_path()),
            sha256sum=bootfile.image_sha256sum,

            host=host, state=state,
        )

    # Synology's SFTP permissions are unusual in that they don't allow
    # you to create directories (which we want to do in the files.template
    # operation after this one). As a workaround to that, we're going to
    # ensure the directory via the files.directory operation since it uses
    # just SSH.
    files.directory(
        name='Ensure grub/ directory exists',
        path=str(host.data.pxe.tftp.root_dir / 'grub'),
        present=True,

        host=host, state=state,
    )

    files.template(
        name='Render GRUB config',
        src=str(templates_base / 'grub.cfg.j2'),
        # files.template uses SFTP to transfer files so we have to use
        # a different base path in the case of Synology which presents a
        # different filesystem hierarchy depending on which protocol you're on.
        # Related bug: https://github.com/Fizzadar/pyinfra/issues/499
        dest=str(host.data.pxe.tftp.sftp_root_dir / 'grub' / 'grub.cfg'),
        create_remote_dir=False,
        pxe=host.data.pxe,
        os_name=str(Path(host.data.pxe.os_image_source_url.path).name),

        host=host, state=state,
    )

    for machine in host.data.pxe.machines:

        machine_ssh_root = host.data.pxe.http.root_dir / machine.hostname
        machine_sftp_root = host.data.pxe.http.sftp_root_dir / machine.hostname

        # Synology's SFTP permissions are unusual in that they don't allow
        # you to create directories (which we want to do in the files.template
        # operations after this one). As a workaround to that, we're going to
        # ensure the directory via the files.directory operation since it uses
        # just SSH.
        files.directory(
            name=f"Ensure {machine_ssh_root} exists",
            path=str(machine_ssh_root),
            present=True,

            host=host, state=state,
        )

        # files.template uses SFTP to transfer files so we have to use
        # a different base path in the case of Synology which presents a
        # different filesystem hierarchy depending on which protocol you're on.
        # Related bug: https://github.com/Fizzadar/pyinfra/issues/499
        meta_data_path = machine_sftp_root / 'meta-data'
        files.template(
            name=f'Render {meta_data_path}',
            src=str(deploy_dir / 'templates' / 'meta-data.j2'),
            dest=str(meta_data_path),
            create_remote_dir=False,
            machine=machine,

            host=host, state=state,
        )

        # files.template uses SFTP to transfer files so we have to use
        # a different base path in the case of Synology which presents a
        # different filesystem hierarchy depending on which protocol you're on.
        # Related bug: https://github.com/Fizzadar/pyinfra/issues/499
        user_data_path = machine_sftp_root / 'user-data'
        files.template(
            name=f'Render {user_data_path}',
            src=str(deploy_dir / 'templates' / 'user-data.j2'),
            dest=str(user_data_path),
            create_remote_dir=False,
            machine=machine,

            host=host, state=state,
        )



from pathlib import Path

from pyinfra.api import deploy
from pyinfra.operations import (
    files,
    server,
)

from aircraft.deploys.ubuntu.models import v1beta2
from aircraft.validators import validate_schema_version

deploy_dir = Path(__file__).parent


@deploy('Configure PXE files')
def configure(state=None, host=None):
    supported_schema_versions = [
        v1beta2.PxeData,
    ]

    validate_schema_version(host.data.pxe, supported_schema_versions)

    #
    # Download bootfile(s) and rener grubg.cfg
    #

    for bootfile in host.data.pxe.bootfiles:
        bootfile_dir = (host.data.pxe.tftp.root_dir / bootfile.get_path()).parent

        files.directory(
            name='Ensure bootfile directory',
            path=str(bootfile_dir),
            present=True,
            sudo=True,

            host=host, state=state,
        )

        files.download(
            name=f'Download bootfile {bootfile.get_path()}',
            src=str(bootfile.image_source_url),
            dest=str(host.data.pxe.tftp.root_dir / bootfile.get_path()),
            sha256sum=bootfile.image_sha256sum,
            sudo=True,

            host=host, state=state,
        )

    files.template(
        name='Render GRUB config',
        src=str(deploy_dir / 'templates' / 'grub.cfg.j2'),
        dest=str(host.data.pxe.tftp.root_dir / 'grub' / 'grub.cfg'),
        pxe=host.data.pxe,
        sudo=True,
        os_name=Path(host.data.pxe.os_image_source_url.path).stem,

        host=host, state=state,
    )

    #
    # Download the OS installer image
    #

    iso_path = \
        host.data.pxe.http.root_dir / host.data.pxe.os_image_source_url.path.lstrip('/')

    files.directory(
        name=f"Ensure {iso_path.parent}",
        path=str(iso_path.parent),
        present=True,
        sudo=True,

        host=host, state=state,
    )

    download_iso = files.download(
        name=f'Download OS Image to {iso_path}',
        src=str(host.data.pxe.os_image_source_url),
        dest=str(iso_path),
        sha256sum=host.data.pxe.os_image_sha256sum,
        sudo=True,

        host=host, state=state,
    )

    #
    # Extract the kernel and ram disk image for use by the bootloader
    #

    kernel_path = str(host.data.pxe.tftp.root_dir / 'vmlinuz')
    initrd_path = str(host.data.pxe.tftp.root_dir / 'initrd')

    if host.fact.file(kernel_path) is None or \
       host.fact.file(initrd_path) is None or \
       download_iso.changed:
        server.shell(
            name='Mount the ISO to /mnt',
            commands=[
                f'mount | grep "{iso_path} on /mnt" || mount {iso_path} /mnt',
            ],
            sudo=True,

            host=host, state=state
        )

        server.shell(
            name="Extract kernel and initrd from ISO",
            commands=[
                f'cp /mnt/casper/vmlinuz {kernel_path}',
                f'cp /mnt/casper/initrd {initrd_path}',
            ],
            sudo=True,

            host=host, state=state,
        )

    server.shell(
        name=f'Ensure {iso_path} is unmounted',
        commands=[
            f'(mount | grep "{iso_path} on /mnt" && umount /mnt) || :',
        ],
        sudo=True,

        host=host, state=state
    )

    #
    # Render the machine-specific user-data and meta-data files
    #

    for machine in host.data.pxe.machines:
        meta_data_path = host.data.pxe.http.root_dir / machine.hostname / 'meta-data'
        files.template(
            name=f'Render {meta_data_path}',
            src=str(deploy_dir / 'templates' / 'meta-data.j2'),
            dest=str(meta_data_path),
            create_remote_dir=True,
            sudo=True,
            machine=machine,

            host=host, state=state,
        )

        user_data_path = host.data.pxe.http.root_dir / machine.hostname / 'user-data'
        files.template(
            name=f'Render {user_data_path}',
            src=str(deploy_dir / 'templates' / 'user-data.j2'),
            dest=str(user_data_path),
            create_remote_dir=True,
            sudo=True,
            machine=machine,

            host=host, state=state,
        )

    # TODO: Check if this should be moved to apache2.py and if doing that requires
    #       that pxe.configure be executed prior to apache2.configure
    files.directory(
        name=f"Ensure www-data has read permissions in {host.data.pxe.http.root_dir}",
        path=str(host.data.pxe.http.root_dir),
        user='root',
        group='www-data',
        mode='0755',
        recursive=True,
        sudo=True,

        state=state, host=host,
    )



from pathlib import Path
from uuid import uuid4

from pydantic import (
    Field,
    validator,
)
from typing import (
    List,
    Optional,
)

from .base_model import BaseModel


class StorageLvmLogicalVolumeConfigData(BaseModel):
    type: str = Field('lvm_partition', const=True)
    id: str = None
    name: str
    size: int
    preserve: bool = False

    format: str
    mount_path: Path

    @validator('id', pre=True, always=True)
    def ensure_id(cls, value):
        return value or f'lv-{uuid4()}'

    def export(self):
        return self.dict(exclude={'schema_version', 'format', 'mount_path'},
                         exclude_none=True)


class StorageLvmVolGroupConfigData(BaseModel):
    type: str = Field('lvm_volgroup', const=True)
    id: str = None
    name: str
    devices: List[str]
    preserve: bool = False

    logical_volumes: List[StorageLvmLogicalVolumeConfigData]

    @validator('id', pre=True, always=True)
    def ensure_id(cls, value):
        return value or f'volgroup-{uuid4()}'

    def export(self):
        return self.dict(exclude={'schema_version', 'logical_volumes'})

    def export_logical_volumes(self):
        return [{**logical_volume.export(), **{'volgroup': self.id}}
                for logical_volume in self.logical_volumes]

    def export_formats(self):
        return [{'fstype': logical_volume.format,
                 'volume': logical_volume.id,
                 'preserve': False,
                 'type': 'format',
                 # id is referenced in self.export_mounts()
                 'id': f'filesystem-for-{logical_volume.id}'}
                for logical_volume in self.logical_volumes]

    def export_mounts(self):
        return [{'device': f'filesystem-for-{logical_volume.id}',
                 'path': str(logical_volume.mount_path),
                 'type': 'mount',
                 'id': f'mount-for-{logical_volume.id}'}
                for logical_volume in self.logical_volumes
                if logical_volume.format is not None]


class StoragePartition(BaseModel):
    type: str = Field('partition', const=True)
    id: str = None
    size: int
    wipe: str = 'superblock'
    flag: str = ''
    preserve: bool = False
    grub_device: Optional[bool]

    format: Optional[str]
    mount_path: Optional[Path]

    @validator('id', pre=True, always=True)
    def ensure_id(cls, value):
        return value or f'partition-{uuid4()}'

    def export(self):
        return self.dict(exclude={'schema_version', 'format', 'mount_path'},
                         exclude_none=True)


class StorageDiskConfigData(BaseModel):
    type: str = Field('disk', const=True)
    id: str = None
    path: str
    ptable: str = 'gpt'
    wipe: str = 'superblock'
    preserve: bool = False
    name: str = ''
    grub_device: bool = False

    partitions: List[StoragePartition]

    @validator('id', pre=True, always=True)
    def ensure_id(cls, value):
        return value or f'disk-{uuid4()}'

    def export(self):
        return self.dict(exclude={'schema_version', 'partitions'})

    def export_partitions(self):
        return [{**partition.export(), **{'device': self.id, 'number': number}}
                for number, partition in enumerate(self.partitions, start=1)]

    def export_formats(self):
        return [{'fstype': partition.format,
                 'volume': partition.id,
                 'preserve': False,
                 'type': 'format',
                 # id is referenced in self.export_mounts()
                 'id': f'filesystem-for-{partition.id}'}
                for partition in self.partitions
                if partition.format is not None]

    def export_mounts(self):
        return [{'device': f'filesystem-for-{partition.id}',
                 'path': str(partition.mount_path),
                 'type': 'mount',
                 'id': f'mount-for-{partition.id}'}
                for partition in self.partitions
                if partition.format is not None]


class StorageConfigData(BaseModel):
    disks: List[StorageDiskConfigData]
    lvm_volgroups: List[StorageLvmVolGroupConfigData]

    def export_disks(self):
        return [disk.export() for disk in self.disks]

    def export_partitions(self):
        return [partition
                for disk in self.disks
                for partition in disk.export_partitions()]

    def export_formats(self):
        return [fmt
                for disk in self.disks
                for fmt in disk.export_formats()] + \
               [fmt
                for volgroup in self.lvm_volgroups
                for fmt in volgroup.export_formats()]

    def export_lvm_volgroups(self):
        return [lvm_volgroup.export() for lvm_volgroup in self.lvm_volgroups]

    def export_lvm_logical_volumes(self):
        return [logical_volume
                for lvm_volgroup in self.lvm_volgroups
                for logical_volume in lvm_volgroup.export_logical_volumes()]

    def export_mounts(self):
        return [mount
                for disk in self.disks
                for mount in disk.export_mounts()] + \
               [mount
                for lvm_volgroup in self.lvm_volgroups
                for mount in lvm_volgroup.export_mounts()]



from ipaddress import (
    IPv4Address,
    IPv4Network,
)
from pydantic import (
    validator,
)
from typing import (
    List,
    Optional,
)

from .base_model import BaseModel
from . import BootfileData


class DhcpRangeData(BaseModel):
    start: IPv4Address
    end: IPv4Address
    lease_time: str = '1m'

    @validator('end')
    def end_must_be_more_than_start(cls, end, values):
        start = values['start']
        if not start <= end:
            raise ValueError(f"DHCP range {start} > {end} is invalid. "
                             "Start address must be less than or equal to end address.")
        return end


class DhcpData(BaseModel):
    subnet: IPv4Network
    ranges: List[DhcpRangeData]
    router: IPv4Address
    dns_servers: List[IPv4Address]
    bootfiles: List[BootfileData]
    tftp_server_name: Optional[IPv4Address]

    @validator('ranges')
    def range_must_be_within_subnet(cls, ranges, values):
        subnet = values['subnet']
        for r in ranges:
            if r.start not in subnet.hosts() or r.end not in subnet.hosts():
                valid_hosts = subnet.hosts()
                first = next(valid_hosts)
                for last in valid_hosts:
                    pass

                raise ValueError(f"DHCP range {r.start} > {r.end} is not within the "
                                 f"subnet {subnet}. Must be within {first} > {last}")
        return ranges



# Always place your charm's dependencies here and not directly in requirements.txt.
# This ensures that when `make dependencies` runs, it will install the runtime
# dependencies correctly. In addition, `make dependencies` will take care of
# updating and pinning the runtime dependencies in requirements.txt so that you
# don't have to. For more info, please see "Adding A Runtime Dependency" in this
# project's README.md file
runtime_requirements = [
    "click",
    "privy",
    "pydantic>=1.7.0,<2.0.0",
    "pyinfra>=1.3.2,<2.0.0",
    "pyyaml",
]

import pathlib
from setuptools import (
    find_packages,
    setup,
)

# Some options here are taken from this handy guide:
# https://realpython.com/pypi-publish-python-package/

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setup(
    install_requires=runtime_requirements,
    version='0.5.2',  # DO NOT EDIT DIRECTLY! Use `bumpversion major|minor|patch`
    name='aircraft',
    author='Mark S. Maglana',
    author_email='mmaglana@gmail.com',
    long_description=README,
    long_description_content_type="text/markdown",
    url='https://github.com/relaxdiego/aircraft',
    python_requires='~=3.6',
    package_dir={'': 'src'},
    include_package_data=True,
    packages=find_packages('src'),
    entry_points={
        'console_scripts': [
            'aircraft = aircraft.cli.main:main'
        ],
    },

    # https://pypi.org/classifiers/
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX :: Linux",
    ]
)



from pathlib import Path

from pyinfra.api import deploy
from pyinfra.operations import (
    apt,
    files,
    systemd,
)

from aircraft.deploys.ubuntu.models import v1beta2
from aircraft.validators import validate_schema_version

deploy_dir = Path(__file__).parent


@deploy('Configure dnsmasq')
def configure(state=None, host=None):
    supported_schemas = [
        v1beta2.DnsmasqData
    ]

    validate_schema_version(host.data.dnsmasq, supported_schemas)

    apt.packages(
        name='Install dnsmasq',
        packages=['dnsmasq'],
        sudo=True,

        state=state, host=host,
    )

    if host.data.dnsmasq.tftp is not None:
        files.directory(
            name=f'Ensure TFTP root dir {host.data.dnsmasq.tftp.root_dir}',
            path=str(host.data.dnsmasq.tftp.root_dir),
            present=True,
            recursive=True,
            sudo=True,

            state=state, host=host,
        )

    dnsmasq_conf = files.template(
        name='Render the dnsmasq config',
        src=str(deploy_dir / 'templates' / 'dnsmasq.conf.j2'),
        dest=str(Path('/etc') / 'dnsmasq.conf'),
        mode='744',
        user='root',
        group='root',
        sudo=True,
        dnsmasq=host.data.dnsmasq,

        state=state, host=host,
    )

    systemd.service(
        name='Restart dnsmasq',
        service='dnsmasq',
        running=True,
        restarted=dnsmasq_conf.changed,
        sudo=True,

        state=state, host=host,
    )


@deploy('Uninstall dnsmasq')
def uninstall(state=None, host=None):
    supported_schemas = [
        v1beta2.DnsmasqData
    ]

    validate_schema_version(host.data.dnsmasq, supported_schemas)

    if 'dnsmasq.service' in host.fact.systemd_status:
        systemd.service(
            name='Stop dnsmasq',
            service='dnsmasq',
            running=False,
            sudo=True,

            state=state, host=host,
        )

    files.file(
        name='Remove dnsmasq config',
        path=str(Path('/etc') / 'dnsmasq.conf'),
        present=False,
        sudo=True,

        state=state, host=host,
    )

    if host.data.dnsmasq.tftp is not None:
        files.directory(
            name=f'Remove TFTP root dir {host.data.dnsmasq.tftp.root_dir}',
            path=str(host.data.dnsmasq.tftp.root_dir),
            present=False,
            recursive=False,
            sudo=True,

            state=state, host=host,
        )

    apt.packages(
        name='Ensure dnsmasq package is not present',
        packages=['dnsmasq'],
        present=False,
        sudo=True,

        state=state, host=host,
    )



from pydantic import (
    AnyHttpUrl,
)
from typing import (
    List,
)

from .base_model import BaseModel
from .bootfile_data import BootfileData
from .machine_data import MachineData
from .tftp_data import TftpData
from .http_data import HttpData


class PxeData(BaseModel):
    tftp: TftpData
    http: HttpData

    os_image_source_url: AnyHttpUrl
    os_image_sha256sum: str

    bootfiles: List[BootfileData]

    machines: List[MachineData]

    preserve_files_on_uninstall: bool = False
